import unittest
import re
import time
from uuid_v9 import uuid, validate_uuid, verify_checksum

uuid_regex = re.compile(r'^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$', re.I)
uuid_v1_regex = re.compile(r'^[0-9a-f]{8}-[0-9a-f]{4}-1[0-9a-f]{3}-[89ab][0-9a-f]{3}-[0-9a-f]{12}$', re.I)
uuid_v4_regex = re.compile(r'^[0-9a-f]{8}-[0-9a-f]{4}-4[0-9a-f]{3}-[89ab][0-9a-f]{3}-[0-9a-f]{12}$', re.I)
uuid_v9_regex = re.compile(r'^[0-9a-f]{8}-[0-9a-f]{4}-9[0-9a-f]{3}-[0-9a-f]{4}-[0-9a-f]{12}$', re.I)

class TestUuidV9(unittest.TestCase):
    def test_validate_as_uuid(self):
        id1 = uuid()
        id2 = uuid('a1b2c3d4')
        id3 = uuid('', False)
        id4 = uuid('a1b2c3d4', False)
        self.assertTrue(uuid_regex.match(id1))
        self.assertTrue(uuid_regex.match(id2))
        self.assertTrue(uuid_regex.match(id3))
        self.assertTrue(uuid_regex.match(id4))

    def test_generate_sequential_ids(self):
        id1 = uuid()
        time.sleep(2)
        id2 = uuid()
        time.sleep(2)
        id3 = uuid()
        self.assertTrue(id1 < id2)
        self.assertTrue(id2 < id3)

    def test_generate_sequential_ids_with_prefix(self):
        id1 = uuid('a1b2c3d4')
        time.sleep(2)
        id2 = uuid('a1b2c3d4')
        time.sleep(2)
        id3 = uuid('a1b2c3d4')
        self.assertTrue(id1 < id2)
        self.assertTrue(id2 < id3)
        self.assertEqual(id1[:8], 'a1b2c3d4')
        self.assertEqual(id2[:8], 'a1b2c3d4')
        self.assertEqual(id3[:8], 'a1b2c3d4')
        self.assertEqual(id1[14:18], id2[14:18])
        self.assertEqual(id2[14:18], id3[14:18])

    def test_generate_non_sequential_ids(self):
        idS = uuid('', False)
        time.sleep(2)
        idNs = uuid('', False)
        self.assertNotEqual(idS[:4], idNs[:4])

    def test_generate_non_sequential_ids_with_prefix(self):
        idS = uuid('a1b2c3d4', False)
        time.sleep(2)
        idNs = uuid('a1b2c3d4', False)
        self.assertEqual(idS[:8], 'a1b2c3d4')
        self.assertEqual(idNs[:8], 'a1b2c3d4')
        self.assertNotEqual(idS[14:18], idNs[14:18])

    def test_generate_ids_with_checksum(self):
        id = uuid('', True, True)
        self.assertTrue(verify_checksum(id))
        self.assertTrue(uuid_regex.match(id))

    def test_generate_ids_with_version(self):
        id = uuid('', True, False, True)
        self.assertTrue(uuid_v9_regex.match(id))

    def test_generate_ids_with_compatibility(self):
        id1 = uuid('', True, False, False, True)
        id2 = uuid('a1b2c3d4', True, False, False, True)
        id3 = uuid('', False, False, False, True)
        id4 = uuid('a1b2c3d4', False, False, False, True)
        self.assertTrue(uuid_v1_regex.match(id1))
        self.assertTrue(uuid_v1_regex.match(id2))
        self.assertTrue(uuid_v4_regex.match(id3))
        self.assertTrue(uuid_v4_regex.match(id4))


    def test_validate_and_verify_checksum(self):
        id1 = uuid('', True, True)
        id2 = uuid('', False, True)
        id3 = uuid('a1b2c3d4', True, True)
        id4 = uuid('a1b2c3d4', False, True)
        id5 = uuid('', True, True, False, True)
        id6 = uuid('', False, True, True, True)
        self.assertTrue(validate_uuid(id1, True))
        self.assertTrue(validate_uuid(id2, True))
        self.assertTrue(validate_uuid(id3, True))
        self.assertTrue(validate_uuid(id4, True))
        self.assertTrue(validate_uuid(id5, True, '1'))
        self.assertTrue(validate_uuid(id6, True, '4'))
        self.assertTrue(verify_checksum(id1))
        self.assertTrue(verify_checksum(id2))
        self.assertTrue(verify_checksum(id3))
        self.assertTrue(verify_checksum(id4))
        self.assertTrue(verify_checksum(id5))
        self.assertTrue(verify_checksum(id6))

if __name__ == '__main__':
    unittest.main()



import re
import random
import time

uuid_regex = re.compile(r'^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$', re.IGNORECASE)

def calc_checksum(hex_string): # CRC-8
    data = [int(hex_string[i:i+2], 16) for i in range(0, len(hex_string), 2)]
    polynomial = 0x07
    crc = 0x00
    for byte in data:
        crc ^= byte
        for _ in range(8):
            if crc & 0x80:
                crc = (crc << 1) ^ polynomial
            else:
                crc <<= 1
    return format(crc & 0xFF, 'x')

def verify_checksum(uuid):
    clean_uuid = uuid.replace('-', '')[0:30]
    checksum = calc_checksum(clean_uuid)
    return checksum == uuid[34:36]

def validate_uuid(uuid, checksum=False, version=False):
    return (isinstance(uuid, str) 
            and uuid_regex.match(uuid)
            and (not checksum or verify_checksum(uuid))
            and (not version or (version is True and uuid[14:15] == '9') or (
                uuid[14:15] == str(version)
                and ('14'.find(str(version)) == -1 or '89abAB'.find(uuid[19:20]) > -1)
            )))

def random_bytes(count):
    return ''.join(random.choice('0123456789abcdef') for _ in range(count))

def random_char(chars):
    random_index = random.randint(0, len(chars) - 1)
    return chars[random_index]

base16_regex = re.compile(r'^[0-9a-fA-F]+$')

def is_base16(str):
    return bool(base16_regex.match(str))

def validate_prefix(prefix):
    if not isinstance(prefix, str):
        raise ValueError('Prefix must be a string')
    if len(prefix) > 8:
        raise ValueError('Prefix must be no more than 8 characters')
    if not is_base16(prefix):
        raise ValueError('Prefix must be only hexadecimal characters')

def add_dashes(str):
    return f'{str[:8]}-{str[8:12]}-{str[12:16]}-{str[16:20]}-{str[20:]}'

def uuid(prefix='', timestamp=True, checksum=False, version=True, compatible=False):
    if prefix:
        validate_prefix(prefix)
        prefix = prefix.lower()
    center = format(int(time.time_ns() / 1000000), 'x') if timestamp is True else format(timestamp, 'x') if isinstance(timestamp, int) else ''
    suffix = random_bytes(32 - len(prefix) - len(center) - (2 if checksum else 0) - (2 if compatible else 1 if version else 0))
    joined = prefix + center + suffix
    if compatible:
        joined = joined[:12] + ('1' if timestamp else '4') + joined[12:15] + random_char('89ab') + joined[15:]
    elif version:
        joined = joined[:12] + '9' + joined[12:]
    if checksum:
        joined += calc_checksum(joined)
    return add_dashes(joined)

if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(description="Generate and validate UUID v9.")
    parser.add_argument("--prefix", dest="prefix", default="", help="Include a prefix (default: '')")
    parser.add_argument("--timestamp", dest="timestamp", help="Customize the timestamp")
    parser.add_argument("--unordered", dest="unordered", action="store_true", help="Exclude timestamp")
    parser.add_argument("--checksum", dest="checksum", action="store_true", help="Include checksum")
    parser.add_argument("--version", dest="version", action="store_true", help="Include version")
    parser.add_argument("--backcompat", dest="backcompat", action="store_true", help="Enable compatibility mode")
    args = parser.parse_args()
    print(uuid(args.prefix, int(args.timestamp) if args.timestamp else not args.unordered, args.checksum, args.version, args.backcompat))